/*
 * Copyright (c) 2003 MATPOCKuH.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <libgen.h>
#include <sysexits.h>
#include <curses.h>
#include <signal.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "settings.h"
#include "acd.h"
#include "logger.h"
#include "version.h"

const char *progname = "acdplay";

int fd = -1;

static volatile sig_atomic_t nobreak = 1;

void
usage(char *argv[])
{
    logger("%s\n\
Usage: %s [-h] [-d device] [-s socket]\n\
  -h\t\t- this help\n\
  -d device\t- device\n\
  -s socket\t- path to a control socket\n",
	ver_getversionstring(), basename(argv[0]));

    exit(EX_USAGE);
}

int
opendev()
{
    int fd;

    if((fd = acd_connect()) == -1) {
	logger("Could not connect to daemon: %s\n", strerror(errno));
	exit(EX_UNAVAILABLE);
    }

    return fd;
}

void
add2data(char **data, char *add)
{
    char *s;

    if(data) {
	asprintf(&s, "%s\n%s", *data, add);
	free(*data);
    } else
	s = strdup(add);

    *data = s;
}

int
execmd(int fd, char *cmd, char **data)
{
    char buf[BUFSIZE], *s, *s_end;
    int i;

    if(!cmd[0])
	return 0;

    *data = NULL;

    asprintf(&s, "%s\n", cmd);
    if(write(fd, s, strlen(s)) < 0) {
	free(s);
	return -1;
    }

    free(s);

    while((i = read(fd, buf, sizeof(buf) - sizeof(char))) > 0) {
	buf[i] = 0;
	s = buf;

	while((s_end = strchr(s, '\n'))) {
	    s_end[0] = 0;

	    if(!strcmp(s, "OK"))
		return 0;

	    if(strstr(s, "ERROR: ") == s) {
		*data = strdup(s);
		return -1;
	    }

	    add2data(data, s);

	    s = s_end + 1;
	}
    }

    if(i < 0)
	return -1;

    return 0;
}

void
mainwindow()
{
    clear();

    mvprintw(0, 0, "  _____________________________________\n\
 |                                     |\n\
 | %s -                      Quit |\n\
 |  __________                         |\n\
 | |          |  Play      Stop        |\n\
 | |          |  SPC pause Eject   +/- |\n\
 | |__________|  Rew       Forward     |\n\
 |_____________________________________|\n", progname);

    doupdate();
}

int
printstatus()
{
    int track, min, sec;
    char *data = NULL, *sb, *se, *status;

    if(execmd(fd, "STATUS", &data) == -1)
	return -1;

    if((status = strchr(data, '<')) == NULL) {
	track = min = sec = 0;
	status = "no cd";
    } else {
	if((++status)[0] == 0)
	    return -1;

	if((se = strchr(status, '>')) == NULL)
	    return -1;

	se++[0] = 0;

	if((sb = strstr(se, " = ")) == NULL)
	    return -1;

	if((se = strchr(sb += 3, ',')) == NULL)
	    return -1;

	se++[0] = 0;
	track = atoi(sb);

	if((sb = strstr(se, " = ")) == NULL)
	    return -1;

	if((se = strchr(sb += 3, ':')) == NULL)
	    return -1;

	se++[0] = 0;
	min = atoi(sb);

	if((sb = strchr(se, '.')) == NULL)
	    return -1;

	sb[0] = 0;
	sec = atoi(se);
    }

    mvprintw(2, 13, "%-9s", status);
    mvprintw(5, 5, "%02i %02i:%02i", track, min, sec);

    free(data);

    move(0, 0);
    doupdate();

    return 0;
}

void
sigterm_handler()
{
    nobreak = 0;
}

int
main(int argc, char *argv[])
{
    int ch;
    char *s, *data;

    while((ch = getopt(argc, argv, "hd:s:")) != -1) {
	switch(ch) {
	case 'd':
	    if(fd == -1)
		fd = opendev();

	    asprintf(&s, "device %s", optarg);
	    if(execmd(fd, s, &data) == -1) {
		logger("%s\n", data);
		free(data);
		free(s);

		return EX_UNAVAILABLE;
	    }

	    free(s);
	    break;

	case 's':
	    acd_setsocketpath(optarg);
	    break;

	case 'h':
	default:
	    usage(argv);
	}
    }

    if(fd == -1)
	fd = opendev();

    if(argc != optind) {
	close(fd);

	usage(argv);
    }

    if(!initscr()) {
	logger("initscr() failed: %s\n", strerror(errno));

	close(fd);

	return EX_UNAVAILABLE;
    }

    cbreak();
    noecho();
    timeout(1000);

    signal(SIGTERM, &sigterm_handler);
    signal(SIGINT, &sigterm_handler);

    mainwindow();

    while(nobreak) {
	char *cmd = NULL;

	if(printstatus() == -1)
	    break;;

	switch(ch = getch()) {
	case 'q':
	case 'Q':
	    nobreak = 0;
	    break;

	case 'p':
	case 'P':
	    cmd = "PLAY";
	    break;

	case ' ':
	    cmd = "PAUSE";
	    break;

	case 'r':
	case 'R':
	    cmd = "PREVIOUS";
	    break;

	case 's':
	case 'S':
	    cmd = "STOP";
	    break;

	case 'e':
	case 'E':
	    cmd = "EJECT";
	    break;

	case 'f':
	case 'F':
	case 'n':
	case 'N':
	    cmd = "NEXT";
	    break;

	case '-':
	    cmd = "VOLUME -10";
	    break;

	case '+':
	    cmd = "VOLUME +10";
	    break;
	}

	if(cmd != NULL)
	    if(execmd(fd, cmd, &data) == -1) {
		free(data);
	    }
    }

    clear();
    refresh();

    endwin();
    close(fd);

    return EX_OK;
}
