/*
 * Copyright (c) 2002, 2003 MATPOCKuH.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <libgen.h>
#include <ctype.h>
#include <sysexits.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_HISTEDIT_H
#include <histedit.h>
#endif

#include "settings.h"
#include "logger.h"
#include "acd.h"
#include "mqueue.h"
#include "version.h"

const char *progname = "cdctl";

typedef struct {
    void *next;
    void *prev;
    char *cmd, *params;
} cmds_t;

mq_queue_t *cmds;

void
usage(char *argv[])
{
    char *scmd, *bname;

    bname = strdup(basename(argv[0]));

    if(!strcmp(bname, "eject") || !strcmp(bname, "pause"))
	scmd = "";
    else {
	scmd = "[command ...]";
    }
	
    logger("%s\n\
Usage: %s [-h] [-d device] [-s socket] %s\n\
  -h\t\t- this help\n\
  -d device\t- device\n\
  -s socket\t- path to a control socket\n\
Use `%s help' for command list\n",
	ver_getversionstring(), bname, scmd, bname);

    free(bname);

    exit(EX_USAGE);
}

int
opendev()
{
    int fd;

    if((fd = acd_connect()) == -1) {
	logger("Could not connect to daemon: %s\n", strerror(errno));
	exit(EX_UNAVAILABLE);
    }

    return fd;
}

void
cleanup(int fd)
{
    close(fd);
    exit(EX_UNAVAILABLE);
}

int
execmd(int fd, char *cmd)
{
    char buf[BUFSIZE], *s, *s_end;
    int i;

    if(!cmd[0])
	return 0;

    asprintf(&s, "%s\n", cmd);

    if(write(fd, s, strlen(s)) < 0) {
	free(s);
	cleanup(fd);
    }

    free(s);

    while((i = read(fd, buf, sizeof(buf) - sizeof(char))) > 0) {
	buf[i] = 0;
	s = buf;

	while((s_end = strchr(s, '\n'))) {
	    s_end[0] = 0;

	    if(!strcmp(s, "OK"))
		return 0;

	    if(strstr(s, "ERROR: ") == s) {
		logger("%s\n", s);
		return -1;
	    }

	    printf("%s\n", s);

	    s = s_end + 1;
	}
    }

    if(i < 0)
	cleanup(fd);

    return 0;
}

int
loadcmds(int fd)
{
    char buf[BUFSIZE], *s, *s_end;
    int i;

    asprintf(&s, "HELP\n");

    if(write(fd, s, strlen(s)) < 0) {
	free(s);
	cleanup(fd);
    }

    free(s);

    cmds = mq_makequeue(sizeof(cmds_t));

    while((i = read(fd, buf, sizeof(buf) - sizeof(char))) > 0) {
	buf[i] = 0;
	s = buf;

	while((s_end = strchr(s, '\n'))) {
	    char *space;
	    cmds_t *tmp;

	    s_end[0] = 0;

	    if(strstr(s, "ERROR: ") == s) {
		logger("Error! Can't load list of avaible commands.\n", s);
		exit(EX_UNAVAILABLE);
	    }

	    if(!strcmp(s, "OK"))
		return 0;

	    tmp = (cmds_t *) mq_addnewrecord(cmds);

	    if((space = strchr(s, ' '))) {
		space[0] = 0;
		tmp->params = strdup(++space);
	    } else {
		tmp->params = NULL;
	    }

	    tmp->cmd = strdup(s);

	    s = s_end + 1;
	}
    }

    if(i < 0)
	cleanup(fd);

    return 0;
}

#ifdef HAVE_HISTEDIT_H
const char *
cdctl_prompt()
{
    return "cdctl> ";
}

char *
input()
{
    static EditLine *el = NULL;
    static History *hist = NULL;
#ifdef EL_INIT_HAVE_STDERR
    HistEvent he;
#endif
    int num = 0, slen;
    const char *bp = NULL;

    for(;;) {
	cmds_t *cmd;
	char *s, *params;

	if (!el) {

#ifdef EL_INIT_HAVE_STDERR
	    el = el_init(progname, stdin, stdout, stderr);
#else
	    el = el_init(progname, stdin, stdout);
#endif

	    hist = history_init();

#ifdef EL_INIT_HAVE_STDERR
	    history(hist, &he, H_SETSIZE, 100);
#else
	    history(hist, H_SETSIZE, 100);
#endif

	    el_set(el, EL_HIST, history, hist);
	    el_set(el, EL_EDITOR, "emacs");
	    el_set(el, EL_PROMPT, cdctl_prompt);
	    el_set(el, EL_SIGNAL, 1);
	    el_source(el, NULL);
	}

	if((bp = el_gets(el, &num)) == NULL || num == 0) {
	    fprintf (stderr, "\r\n");
	    return NULL;
	}

#ifdef EL_INIT_HAVE_STDERR
	history(hist, &he, H_ENTER, bp);
#else
	history(hist, H_ENTER, bp);
#endif

	if(bp[0] == 0 || bp[0] == '\n') {
	    continue;
	}

	if((s = strdup(bp))[(slen = strlen(bp)) - 1] == '\n')
	    s[--slen] = 0;

	if((params = strchr(s, ' '))) {
	    params[0] = 0;
	    slen = strlen(s);
	}
	    
	for(mq_resetcurrent(cmds); (cmd = (cmds_t *) mq_nextrecord(cmds));) {
	    if(params) {
		if(cmd->params) {
		    if(!strncasecmp(s, cmd->cmd, slen)) {
			params[0] = ' ';
			return s;
		    }
		}
	    } else {
		if(!cmd->params) {
		    if(!strncasecmp(s, cmd->cmd, slen)) {
			return s;
		    }
		}
	    }
	}

	logger("invalid command, enter ``help'' for commands\n");
    }
}
#endif

int
main(int argc, char *argv[])
{
    int ch, fd = -1;
    char *bname, *s;

    while((ch = getopt(argc, argv, "hd:s:")) != -1) {
	switch(ch) {
	case 'd':
	    if(fd == -1)
		fd = opendev();

	    asprintf(&s, "device %s", optarg);
	    if(execmd(fd, s) == -1) {
		free(s);
		return EX_UNAVAILABLE;
	    }

	    free(s);
	    break;

	case 's':
	    acd_setsocketpath(optarg);
	    break;

	case 'h':
	default:
	    usage(argv);
	}
    }

    if(fd == -1)
	fd = opendev();

    bname = strdup(basename(argv[0]));

    if(!strcmp(bname, "eject") || !strcmp(bname, "pause")) {
	if(argc != optind)
	    usage(argv);

	execmd(fd, bname);

	close(fd);
	return EX_OK;
    }

    loadcmds(fd);

    if(argc == optind) {
#ifdef HAVE_HISTEDIT_H
	char *command;

	printf("%s\nType ``help'' for command list\n\n",
	    ver_getversionstring());

	while((command = input())) {
	    int slen = strlen(command);

	    if(command[slen - 1] == '\n')
		command[--slen] = 0;

	    execmd(fd, command);

	    if(command[0] && !strncasecmp(command, "QUIT", strlen(command))) {
		free(command);
		break;
	    }

	    free(command);
	}

	close(fd);

	return EX_OK;
#else
	close(fd);
	usage(argv);
#endif
    }

    for(; optind < argc; optind++) {
	cmds_t *cmd;
	int paramc = 0, slen, notexecuted = 1;
	char *command;

	slen = strlen(argv[optind]);

	if((optind + 1) < argc) {
	    if(isdigit(argv[optind + 1][0]) ||
		argv[optind + 1][0] == '-' || argv[optind + 1][0] == '+') {
		if((optind + 2) < argc)
		    if(isdigit(argv[optind + 2][0]))
		    	paramc = 2;
		    else
			paramc = 1;
		else
		    paramc = 1;
	    }
	}

	switch(paramc) {
	case 0:
	    asprintf(&command, "%s", argv[optind]);
	    break;

	case 1:
	    asprintf(&command, "%s %s", argv[optind], argv[optind + 1]);
	    break;

	case 2:
	    asprintf(&command, "%s %s %s",
		argv[optind], argv[optind + 1], argv[optind + 2]);
	    break;
	}

	for(mq_resetcurrent(cmds); (cmd = (cmds_t *) mq_nextrecord(cmds));) {
	    if(paramc == 0) {
		if(!cmd->params) {
		    if(!strncasecmp(argv[optind], cmd->cmd, slen)) {
			execmd(fd, command);
			notexecuted = 0;
			break;
		    }
		}
	    } else {
		if(cmd->params) {
		    if(!strncasecmp(argv[optind], cmd->cmd, slen)) {
			execmd(fd, command);
			notexecuted = 0;
			break;
		    }
		}
	    }
	}

	if(notexecuted)
	    logger("Unknown command: %s\n", argv[optind]);
	else	
	    optind += paramc;
    }

    close(fd);

    return EX_OK;
}
